from dash import html


def navbar():
    """Navbar component

    Returns:
        dash.html.Nav: Navbar component
    """
    return html.Nav(
        [
            html.Div(
                html.A(
                    html.Img(
                        src="../assets/logos/SQUAD Logo.webp",
                        alt="logo",
                        id="logo",
                        style={"maxHeight": "90%", "maxWidth": "100%"},
                    ),
                    href="https://squad-lab.org",
                    className="navbar-item",
                ),
                className="navbar-brand",
            ),
            html.Div(
                html.A("QIMCHI", href="./", className="navbar-item"),
                className="navbar-end",
            ),
        ],
        className="navbar is-transparent",
        id="navbar",
        **{"data-theme": "light"},
    )



"""
Author: Spandan Anupam
Affiliation: Forschungszentrum Jülich GmbH

"""

from dash import Dash
from starlette.applications import Starlette
from starlette.middleware.wsgi import WSGIMiddleware


external_scripts = []
external_stylesheets = [
    "https://cdn.jsdelivr.net/npm/bulma@1.0.0/css/bulma.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.5.2/css/all.min.css",
]

app = Dash(
    __name__,
    assets_folder="./assets",
    external_scripts=external_scripts,
    external_stylesheets=external_stylesheets,
    use_pages=True,
    title="Qimchi",
    update_title=None,
    suppress_callback_exceptions=True,
)

"""
Entry point for running the Starlette app with Dash.

"""
server = app.server

asgi_app = Starlette()
asgi_app.mount("/", WSGIMiddleware(server))



from pathlib import Path
from typing import Any, List

import dash
from dash import Input, Output, State, callback, dcc, html
from dash.exceptions import PreventUpdate

# Local imports
from ..base import Data, Database
from qimchi.state import _state
from qimchi.logger import logger
from qimchi.components.utils import read_data


def _update_options_fn(
    condition_list: List[str | Path | Any],
    check_file: bool = False,
    default_return: List[str] | None = [],
) -> List[str] | str | None:
    """
    Helper function to update options based on the condition list.

    Args:
        condition_list (List[str | Path | Any]): List of parameters to compare with None.
            Ordering matters, as the path is constructed from the first item to the second last item.
            The last item is always assumed to be `dataset_type`
        check_file (bool, optional): Check if the path is a file, by default False (checks for directory)
        default_return (List | None, optional): Default return item if the condition is not satisfied, by default []

    Returns:
        List[str] | str | None: List of strings (options) or empty list or None

    """
    # NOTE: This only checks for truthy (i.e., "" is False, but Path("") is True). Should work here though.
    if all(condition_list):
        # Assuming either strings or Paths
        path = Path(*condition_list[:-1])

        if not path.is_dir():
            logger.debug(f"Path {path} is not a directory.")
            return default_return
        # Unfiltered options
        options: List[str] = [p.name for p in path.iterdir()]
        # Filter for only files or directories
        options = [
            opt
            for opt in options
            if (
                # TODOLATER: Some redundancy here
                (path / opt).is_dir() and (path / opt).suffix == ".zarr"
                if check_file
                else (path / opt).is_dir()
            )
        ]
        return options

    return default_return


def _should_update_data(path: Path, sig_curr: int, origin: str) -> bool:
    """
    Conditionally update the data store with the Zarr file content.

    Args:
        path (Path): Path to the dataset
        sig_curr (int): The current signal

    Returns:
        bool: Whether the data has been updated

    """
    if path.is_dir() and path.suffix == ".zarr":
        curr_mod_time = path.stat().st_mtime
        if path == _state.measurement_path:
            if curr_mod_time != _state.measurement_last_fmt:
                logger.warning(
                    f"_should_update_data | {origin}: DATA UPDATED ON DISK!! AT {path}."
                )
                _state.measurement_last_fmt = curr_mod_time
                _state.save_state()
                return True

            # If the data has not been loaded even once, load it (e.g., for the first time on page load)
            if sig_curr in [None, 0]:
                logger.warning(
                    f"_should_update_data | {origin}: DATA BEING LOADED FROM {path}"
                )
                return True
            # If the data has been loaded, do not update
            else:
                return False

        else:
            # If the path has changed, update the state
            logger.warning(
                f"_should_update_data | {origin}: DATA PATH MODIFIED!! AT {path}."
            )
            _state.measurement_path = path
            _state.measurement_last_fmt = curr_mod_time
            _state.save_state()

            # logger.warning(f"{origin}: LOADING DATA FROM MODIFIED {path}")
            # ds = xr.load_dataset(path, engine="zarr")  # TODONOW: .to_dict()

            if origin == "XarrayData":
                # To update the wafer_id etc. in the state
                data = read_data()
                metadata = data.attrs
                device_type = metadata.get("Device Type", "")
                wafer_id = metadata.get("Wafer ID", "")
                sample_name = metadata.get("Sample Name", "")
                meas_id = metadata.get("Measurement ID", "")

                # Update the state
                _state.device_type = device_type
                _state.wafer_id = wafer_id
                _state.device_id = sample_name
                _state.measurement = meas_id
                _state.save_state()

            logger.warning(
                f"_should_update_data | {origin}: DATA LOADED FROM MODIFIED {path}"
            )
            return True


class XarrayDataFolder(Database):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    @callback(
        Input("measurement", "value"),
        Input("measurement-type", "value"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
    )
    def update_state_wafer_id(
        measurement: str,
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
    ) -> None:
        """
        Update the state with the wafer id, device type, device id, measurement type, and measurement.

        """
        _state.wafer_id = wafer_id if wafer_id else _state.wafer_id
        _state.device_type = device_type if device_type else _state.device_type
        _state.device_id = device_id if device_id else _state.device_id
        _state.measurement_type = (
            measurement_type if measurement_type else _state.measurement_type
        )
        _state.measurement = measurement if measurement else _state.measurement
        _state.save_state()

    @staticmethod
    @callback(
        Output("wafer-id", "options"),
        State("wafer-id", "options"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("selector", "children"),
        Input("metadata-viewer", "children"),
        prevent_initial_call=True,
    )
    def update_wafer_id(
        wafer_options: List[str], dataset_path: str | Path, dataset_type: str, *_
    ) -> List[str]:
        return _update_options_fn(
            [dataset_path, dataset_type], default_return=wafer_options
        )

    @staticmethod
    @callback(
        Output("device-type", "options"),
        State("device-type", "options"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("selector", "children"),
        Input("metadata-viewer", "children"),
        prevent_initial_call=True,
    )
    def update_device_type(
        device_type_options: List[str],
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        return _update_options_fn([dataset_path, wafer_id, dataset_type])

    @staticmethod
    @callback(
        Output("device-id", "options"),
        State("device-id", "options"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("selector", "children"),
        Input("metadata-viewer", "children"),
        prevent_initial_call=True,
    )
    def update_device_id(
        device_id_options: List[str],
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        return _update_options_fn([dataset_path, wafer_id, device_type, dataset_type])

    @staticmethod
    @callback(
        Output("measurement-type", "options"),
        State("measurement-type", "options"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("selector", "children"),
        Input("metadata-viewer", "children"),
        prevent_initial_call=True,
    )
    def update_measurement_type(
        measurement_type_options: List[str],
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        return _update_options_fn(
            [dataset_path, wafer_id, device_type, device_id, dataset_type],
        )

    @staticmethod
    @callback(
        Output("measurement", "options"),
        State("measurement", "options"),
        Input("measurement-type", "value"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("selector", "children"),
        Input("metadata-viewer", "children"),
        prevent_initial_call=True,
    )
    def update_measurement(
        measurement_options: List[str],
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        options = _update_options_fn(
            [
                dataset_path,
                wafer_id,
                device_type,
                device_id,
                measurement_type,
                dataset_type,
            ],
            check_file=True,  # Checks if the path is a measurement file
        )
        options.sort(key=lambda x: int(x.split("-")[0]))
        return options

    @staticmethod
    @callback(
        Output(
            "load-signal",
            "data",
            allow_duplicate=True,
        ),
        State("load-signal", "data"),
        Input(
            "measurement",
            "value",
        ),
        State("measurement-type", "value"),
        State("device-id", "value"),
        State("device-type", "value"),
        State("wafer-id", "value"),
        State("dataset-path", "value"),
        Input("selector", "children"),
        Input("upload-ticker", "n_intervals"),
        prevent_initial_call=True,
    )
    def update_data(
        sig_curr: int,
        measurement: str,
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        *_,
    ) -> dict:
        """
        Update the data store with the Zarr file content.

        Args:
            sig_curr (int): The current signal value
            measurement (str): Measurement name
            measurement_type (str): Measurement type
            device_id (str): Device ID
            device_type (str): Device type
            wafer_id (str): Wafer ID
            dataset_path (str | Path): Path to the dataset

        Returns:
            dict: The updated data store

        Raises:
            PreventUpdate: If the path is not a file or not a Zarr file

        """
        if measurement is None:
            raise PreventUpdate

        if all(
            [
                measurement,
                measurement_type,
                device_id,
                device_type,
                wafer_id,
                dataset_path,
            ]
        ):
            measurement_path = Path(
                dataset_path,
                wafer_id,
                device_type,
                device_id,
                measurement_type,
                measurement,
            )
            path = measurement_path
            if _should_update_data(path, sig_curr, "XarrayDataFolder"):
                return sig_curr + 1 if sig_curr is not None else 1
            else:
                raise PreventUpdate

    @staticmethod
    @callback(
        Output("measurement", "value"),
        Input("prev-mmnt-button", "n_clicks"),
        Input("next-mmnt-button", "n_clicks"),
        Input("measurement", "value"),
        State("measurement", "options"),
    )
    def update_selected_measurement(
        _prev_clicks: int, _next_clicks: int, curr_mmnt: str, mmnt_options: list
    ) -> int:
        """
        Updates selected measurement based on button clicks or dropdown selection.

        Args:
            _prev_clicks (int): Number of clicks on the previous button
            _next_clicks (int): Number of clicks on the next button
            curr_mmnt (str): Current selected measurement
            mmnt_options (list): List of measurement options

        Returns:
            str: New selected measurement

        Raises:
            PreventUpdate: If the options are not available

        """
        if mmnt_options is None:
            old_mmnt = _state.measurement
            # NOTE: Assuming this is None only on reload
            logger.debug(
                f"Options not available: Returning old measurement: {old_mmnt}"
            )
            return old_mmnt

        # Identify which button was pressed
        triggered_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        curr_idx = mmnt_options.index(curr_mmnt)
        if triggered_id == "prev-mmnt-button":
            new_index = (curr_idx - 1) % len(mmnt_options)  # Cycle left
        elif triggered_id == "next-mmnt-button":
            new_index = (curr_idx + 1) % len(mmnt_options)  # Cycle right
        else:
            new_index = curr_idx  # No change

        logger.debug(f"New index: {new_index}")
        logger.debug(f"New option: {mmnt_options[new_index]}")

        return mmnt_options[new_index]

    def options(self):
        return html.Div(
            [
                html.Div("Sample Information:", className="column is-1"),
                html.Div(
                    [
                        dcc.Dropdown(
                            id="wafer-id",
                            placeholder="Wafer ID",
                            searchable=True,
                            persistence=True,
                            persistence_type="local",
                        ),
                    ],
                    className="column is-2",
                ),
                html.Div(
                    [
                        dcc.Dropdown(
                            id="device-type",
                            placeholder="Device Type",
                            searchable=True,
                            persistence=True,
                            persistence_type="local",
                        ),
                    ],
                    className="column is-2",
                ),
                html.Div(
                    [
                        dcc.Dropdown(
                            id="device-id",
                            placeholder="Device ID",
                            searchable=True,
                            persistence=True,
                            persistence_type="local",
                        ),
                    ],
                    className="column is-2",
                ),
                html.Div(
                    [
                        dcc.Dropdown(
                            id="measurement-type",
                            placeholder="Measurement Type",
                            searchable=True,
                            persistence=True,
                            persistence_type="local",
                        ),
                    ],
                    className="column is-2",
                ),
                html.Div(
                    [
                        dcc.Dropdown(
                            id="measurement",
                            placeholder="Measurement",
                            searchable=True,
                            # NOTE: Persistence handled server-side
                        )
                    ],
                    className="column",
                ),
                html.Button(
                    html.I(className="fa-solid fa-arrow-left button"),
                    id="prev-mmnt-button",
                    n_clicks=0,
                ),
                html.Button(
                    html.I(className="fa-solid fa-arrow-right button"),
                    id="next-mmnt-button",
                    n_clicks=0,
                ),
            ],
            className="columns is-full is-multiline is-flex is-vcentered",
        )


class XarrayData(Data):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    @callback(
        Output(
            "load-signal",
            "data",
            allow_duplicate=True,
        ),
        State("load-signal", "data"),
        Input("dataset-path", "value"),
        Input("upload-ticker", "n_intervals"),
        prevent_initial_call=True,
    )
    def update_data(sig_curr: int, path: str, *_):
        """
        Update the data store with the Zarr file content.

        Args:
            sig_curr (int): The current signal value
            path (str): Path to the Zarr file

        Returns:
            dict: The updated data store

        Raises:
            PreventUpdate: If the path is not a file or not a Zarr file

        """
        if path is None:
            raise PreventUpdate

        path = Path(path)
        if _should_update_data(path, sig_curr, "XarrayData"):
            return sig_curr + 1 if sig_curr is not None else 1
        else:
            raise PreventUpdate



from pathlib import Path
import xarray as xr
from dash.exceptions import PreventUpdate

# Local imports
from qimchi.state import _state
from qimchi.logger import logger


def _retry_reader(path, retries=3):
    """
    Utility function to retry reading a dataset

    Args:
        path (Path): Path to the dataset
        retries (int, optional): Number of retries. Defaults to 3.

    Returns:
        xr.Dataset: Loaded xarray dataset

    """
    for i in range(retries):
        try:
            ds = xr.load_dataset(path, engine="zarr")
            return ds
        except Exception as err:
            logger.warning(f"_retry_reader | Error reading: {path} | err: {err}")
            logger.warning(f"_retry_reader | Retrying... {i + 1}/{retries}")

    logger.error(f"_retry_reader | Failed to read: {path} after {retries} retries")
    raise PreventUpdate


def read_data() -> xr.Dataset:
    """
    Utility function to read data into a xarray.Dataset

    Returns:
        xr.Dataset: Loaded xarray dataset

    """
    path = Path(_state.measurement_path)
    logger.debug(f"read_data | Reading: {path}")

    try:
        ds = xr.load_dataset(path, engine="zarr")
        logger.debug(f"read_data | Reading done: {path}")
        return ds
    # Catch "FileNotFoundError: Unable to find group" and keep retrying
    except FileNotFoundError as err:
        logger.warning(f"read_data | FileNotFoundError reading: {path} | err: {err}")
        logger.warning("read_data | Retrying...")
        ds = _retry_reader(path)
        return ds
    except KeyError as err:
        logger.warning(f"read_data | KeyError reading: {path} | err: {err}")
        logger.warning("read_data | Retrying...")
        ds = _retry_reader(path)
        return ds
    except Exception as err:
        logger.error(f"read_data | Error reading: {path} | err: {err}", exc_info=True)
        raise PreventUpdate



from pathlib import Path

from dash import Input, Output, State, callback, dcc, html
from dash.exceptions import PreventUpdate

# Local imports
import qimchi.components.data as data
from qimchi.components.data import (
    __all__ as DATASET_TYPES,
)  # NOTE: __all__ is a list of all the public names in the module
from qimchi.components.utils import read_data
from qimchi.state import _state, DATA_REFRESH_INTERVAL
from qimchi.logger import logger


def data_selector() -> html.Div:
    """
    Generator for the data selector component.

    Returns:
        dash.html.Div: The data selector component

    """
    return html.Div(
        [
            html.Div(
                [
                    dcc.Interval(
                        id="upload-ticker",
                        interval=DATA_REFRESH_INTERVAL,
                        n_intervals=0,
                    ),
                    html.Div(
                        dcc.Input(
                            className="input",
                            type="text",
                            placeholder="Dataset Path",
                            id="dataset-path",
                            persistence=True,
                            persistence_type="local",
                        ),
                        className="column is-5 mb-0 pb-0",
                    ),
                    html.Div(
                        dcc.Dropdown(
                            options=DATASET_TYPES,
                            placeholder="Dataset type",
                            id="dataset-type",
                            persistence=True,
                            persistence_type="local",
                        ),
                        className="column is-2 mb-0 pb-0",
                    ),
                    html.Div(
                        html.Button(
                            "Submit", id="submit", className="button is-warning"
                        ),
                        className="column is-2 mb-0 pb-0",
                    ),
                    html.Div(
                        className="column is-12 m-1 pt-0 mt-0",
                        id="data-options",
                    ),
                    dcc.Store("load-signal", data=0),
                    # TODONOW:
                    dcc.Store("load-end-signal", data=0),
                ],
                className="columns is-full is-multiline ml-1 mr-1 is-flex is-vcentered",
                id="selector",
            ),
        ]
    )


@callback(
    Output("data-options", "children"),
    State("data-options", "children"),
    Input("dataset-type", "value"),
    Input("dataset-path", "value"),
    Input("submit", "n_clicks"),
    prevent_initial_call=True,
)
def update_options(
    sel_contents: None | html.Div, dataset_type: str, dataset_path: str, _
) -> html.Div:
    """
    Updates the options for the data selector.

    Args:
        sel_contents (None | html.Div): The current contents of the data selector
        dataset_type (str): The type of the dataset
        dataset_path (str): The path to the dataset

    Returns:
        dash.html.Div: The updated data selector component

    """
    if dataset_type is not None and dataset_path is not None:
        try:
            dataset_path = Path(dataset_path)
            logger.debug(f"Dataset Type: {dataset_type}")
            logger.debug(f"Dataset Path: {dataset_path}")

            # Import `dataset_type` class from data module and instantiate it
            data_cls = getattr(data, dataset_type)(path=dataset_path)
            logger.debug(f"Dataset Class: {data_cls}")

            # Update the state
            _state.dataset_path = dataset_path
            _state.dataset_type = dataset_type
            _state.save_state()

            return data_cls.selector()
        except AttributeError:
            # CONCERN: API: XarrayData is being handled differently from XarrayDataFolder
            logger.error("AttributeError from update_options()")
            return sel_contents
    return sel_contents


@callback(
    Output("submit", "n_clicks"),
    State("submit", "n_clicks"),
    State("data-options", "children"),
    Input("upload-ticker", "n_intervals"),
)
def refresh(n_clicks: int, data_options: list, _) -> int:
    """
    Refreshes the submit button, auto-submitting the data path and options.

    Args:
        n_clicks (int): The current number of clicks
        data_options (list): The data options

    Returns:
        int: The number of clicks

    """
    if data_options is None:
        return n_clicks
    else:
        raise PreventUpdate


@callback(
    Output("dependent-dropdown", "options"),
    Input("load-signal", "data"),
)
def update_dependents(sig: int | None) -> list:
    """
    Updates the dependent dropdown options.

    Args:
        sig (int | None): Signal to indicate that data has been updated

    Returns:
        list: The dependent dropdown options generated from the data

    Raises:
        PreventUpdate: If `sig` is None or 0

    """
    if sig in [None, 0]:
        raise PreventUpdate
    data = read_data()
    return list(data.data_vars.keys())


@callback(
    Output("independent-dropdown", "options"),
    Input("load-signal", "data"),
    Input("dependent-dropdown", "value"),
)
def update_independents(sig: int | None, dependents: list):
    """
    Updates the independent dropdown options.

    Args:
        sig (int): Signal to indicate that data has been updated
        dependents (list): List of dependent variables

    Returns:
        list: The independent dropdown options generated from the data

    Raises:
        PreventUpdate: If `sig` is None or 0, or `dependents` is None

    """
    if sig in [None, 0] or dependents is None:
        raise PreventUpdate
    data = read_data()
    return list(data[dependents].coords)
